#!/usr/bin/env python

#    Copyright (C) 2011, Paul Becker
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pyglet, os, random, time, math
from pyglet import gl
from pyglet.window import key as KEY


def strtosprites(text, x, y, scale=1):
	lst = []
	for c in text:
		if c in GAME.images:
			s = pyglet.sprite.Sprite(GAME.images[c],
				batch=GAME.menubatch)
			s.x = x
			s.y = y
			s.scale = scale
			s.color = (255, 255, 255)
			s.opacity = 191
			lst.append(s)
		x += round(8 * scale)
	return lst


class Menu(object):
	def __init__(self, status=''):
		self.status = status
		self.infosprites = []
		self.help = True
		self.togglehelp()
		self.titlesprites = strtosprites('crabhack', -59, -43, scale=2)
		
		pyglet.clock.schedule_interval(self.tick, .75)
		GAME.window.push_handlers(self)
		GAME.dim = True
	
	def togglehelp(self):
		self.help = not self.help
		if self.help:
			if GAME.stats:
				GAME.stats.setvis(False)
			self.setinfo(
				'baby', 'use hjklyubn to move\nor arrows or numpad',
				'fish', 'eat fish to grow',
				'bird', 'avoid birds and other\ncrabs until you mature',
				'crab', 'once you are an adult\nfind a crab to mate',
				'baby', 'continue playing\nas the baby crab',
				'press enter to begin')
		else:
			if GAME.stats:
				GAME.stats.setvis(True)
			self.setinfo(
				'',
				self.status,
				'',
				'press h for help',
				'press enter to begin')
	
	def setinfo(self, *lst):
		for s in self.infosprites:
			s.delete()
		self.infosprites = []
		x, y = -59, 40
		h = 3
		for s in lst:
			if s in GAME.images:
				s = pyglet.sprite.Sprite(GAME.images[s], batch=GAME.menubatch)
				s.x = x
				s.y = y
				self.infosprites.append(s)
				x += int(s.width * 1.5)
				h = s.height / 2
			else:
				for line in s.split('\n'):
					self.infosprites += strtosprites(line, x, y + h, scale=.6)
					h -= 5
				x = -59
				y -= 12
	
	def close(self):
		pyglet.clock.unschedule(self.tick)
		GAME.window.pop_handlers()
		GAME.dim = False
		for s in self.infosprites + self.titlesprites:
			s.delete()
	
	def tick(self, dt):
		GAME.updateworld()


class Stats:
	def __init__(self):
		self.fish = 0
		self.steps = 0
		self.lives = 1
		self.sprites = []
		self.update()
	
	def update(self):
		if not GAME.pos:
			return
		for s in self.sprites:
			s.delete()
		self.sprites = []
		y = GAME.pos[1]
		ns = ' '
		if y < 0:
			y = -y
			ns = 's'
		elif y > 0:
			ns = 'n'
		if self.lives != 1:
			life = 'lives'
		else:
			life = 'life'
		self.sprites += strtosprites(
			'%d fish  %d %s  %d %s'
				% (self.fish, y, ns, self.lives, life),
			-59, 48, scale=.6)
	
	def setvis(self, v):
		if v: a = 191
		else: a = 0
		for s in self.sprites:
			s.opacity = a


class Game:
	def __init__(self):
		self.window = pyglet.window.Window(640, 480)
		self.keys = KEY.KeyStateHandler()
		self.window.push_handlers(self.keys)
		self.window.push_handlers(self)
		
		gl.glEnable(gl.GL_BLEND)
		gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
		
		self.dim = False
		self.scrverts = ('v2f', [
			-self.window.width/2, self.window.height/2,
			self.window.width/2, self.window.height/2,
			self.window.width/2, -self.window.height/2,
			-self.window.width/2, -self.window.height/2,
		])
		
		self.stats = None
		self.coastframe = 0
		self.coastx = 6
		self.rockx = -15
		
		files = ['images.png', 'images.txt']
		image = None
		for path in ['/opt/crabhack', '/usr/share/games/crabhack', os.path.dirname(os.path.abspath(__file__)), '']:
			count = 0
			for f in files:
				if os.path.exists(os.path.join(path, f)):
					count += 1
			if count == len(files):
				image = pyglet.image.load(os.path.join(path, 'images.png'))
				imagenames = open(os.path.join(path, 'images.txt'), 'r').read().split()
				break
		if not image:
			raise Exception('failed to find resource files')
		
		size = int(math.sqrt(len(imagenames))) + 1
		self.atlas = pyglet.image.atlas.TextureAtlas(size * 10, size * 10)
		self.images = {}
		for i in range(len(imagenames)):
			x = (i % size) * 10
			y = (size - 1 - i / size) * 10
			self.images[imagenames[i]] = self.atlas.add(image.get_region(x, y, 10, 10))
		t = self.images[self.images.keys()[0]].texture
		gl.glBindTexture(t.target, t.id)
		gl.glTexParameteri(t.target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
		
		self.batch = pyglet.graphics.Batch()
		self.menubatch = pyglet.graphics.Batch()
		self.gamesprites = {}
		for y in range(-5, 6):
			for x in range(-6, 7):
				s = pyglet.sprite.Sprite(self.images['sand'])
				s.x = x * 10
				s.y = y * 10
				s.batch = self.batch
				self.gamesprites[(x, y)] = s
		
		self.clear(False)
	
	def clear(self, player):
		self.fishcount = 0
		if player:
			self.board = {(0, 0): 'baby'}  # {(x, y): img}
			self.pos = (0, 0)
			self.cam = (0, 0)
			self.stats = Stats()
		else:
			self.board = {}
			self.pos = None
			self.cam = (2, 0)
		self.movers = []  # [(x, y)]
		
		self.genmin = -10
		self.genmax = 10
		for y in range(self.genmin, self.genmax + 1):
			self.genline(y)
	
	def genline(self, y):
		for x in range(self.rockx + 1, self.rockx + 10):
			if random.random() > (x - self.rockx) / 10.:
				self.board[(x, y)] = 'rock'
				if (x, y) in self.movers: self.movers.remove((x, y))
		c = random.choice([0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 3])
		while c:
			x = random.randint(self.rockx + 1, self.coastx - 1)
			if (x, y) in self.board: continue
			r = random.random()
			if r < .5:
				self.board[(x, y)] = 'rock'
			elif r < .8:
				self.board[(x, y)] = 'bird'
				self.movers.append((x, y))
			elif r < .9:
				self.board[(x, y)] = 'fish'
			else:
				self.board[(x, y)] = 'crab'
				self.movers.append((x, y))
			c -= 1
	
	def on_draw(self):
		for dx in range(-6, 7):
			for dy in range(-5, 6):
				s = self.gamesprites[(dx, dy)]
				x, y = dx + self.cam[0], dy + self.cam[1]
				
				if (x, y) in self.board:
					img = self.board[(x, y)]
					if img.startswith('crab') and img != 'crab':
						img = 'crab' + str((int(img[4:]) % 2) + 1)
				elif x == self.coastx: img = 'coast' + str(self.coastframe)
				elif x > self.coastx: img = 'ocean'
				elif x <= self.rockx: img = 'rock'
				else: img = 'sand'
				
				i = self.images[img]
				if s.image != i: s.image = i
		
		gl.glLoadIdentity()
		gl.glTranslated(self.window.width / 2 - 25, self.window.height / 2 - 25, 0)
		gl.glScaled(5, 5, 0)
		self.batch.draw()
		
		if self.dim:
			gl.glColor4f(0, 0, 0, .5)
			pyglet.graphics.draw(4, gl.GL_QUADS, self.scrverts)
		
		if self.stats:
			self.stats.update()
		
		self.menubatch.draw()
	
	def on_key_press(self, key, mods):
		if key == KEY.H and not self.pos:
			MENU.togglehelp()
		elif key in [KEY.H, KEY.LEFT, KEY.NUM_4, KEY.NUM_LEFT]:
			self.move(-1, 0)
		elif key in [KEY.J, KEY.DOWN, KEY.NUM_2, KEY.NUM_DOWN]:
			self.move(0, -1)
		elif key in [KEY.K, KEY.UP, KEY.NUM_8, KEY.NUM_UP]:
			self.move(0, 1)
		elif key in [KEY.L, KEY.RIGHT, KEY.NUM_6, KEY.NUM_RIGHT]:
			self.move(1, 0)
		elif key in [KEY.Y, KEY.NUM_7, KEY.NUM_HOME]:
			self.move(-1, 1)
		elif key in [KEY.U, KEY.NUM_9, KEY.NUM_PAGE_UP]:
			self.move(1, 1)
		elif key in [KEY.B, KEY.NUM_1, KEY.NUM_END]:
			self.move(-1, -1)
		elif key in [KEY.N, KEY.NUM_3, KEY.NUM_PAGE_DOWN]:
			self.move(1, -1)
		elif key in [KEY.PERIOD, KEY.NUM_5, KEY.NUM_BEGIN]:
			self.move(0, 0)
		elif key == KEY.ENTER and not self.pos:
			MENU.close()
			self.clear(True)
	
	def move(self, dx, dy):
		if not self.pos: return
		np = (self.pos[0] + dx, self.pos[1] + dy)
		if np not in self.board:
			if np[0] <= self.rockx or np[0] >= self.coastx:
				np = None
		elif self.board[np] == 'crab':
			if self.board[self.pos] == 'crab':
				self.board[np] = 'crab10'
				self.board[self.pos] = 'crab11'
				self.movers.append(self.pos)
				lst = []
				for pos in [np, self.pos]:
					for dx in range(-1, 2):
						for dy in range(-1, 2):
							p = (pos[0] + dx, pos[1] + dy)
							if p not in self.board:
								lst.append(p)
				self.pos = random.choice(lst)
				self.board[self.pos] = 'baby'
				self.fishcount = 0
				self.stats.lives += 1
			np = None
		elif self.board[np] == 'fish':
			self.fishcount += 1
			self.stats.fish += 1
			if self.board[self.pos] == 'baby' and self.fishcount >= 5:
				self.board[self.pos] = 'crab'
		else: np = None
		
		if np:
			self.board[np] = self.board[self.pos]
			del self.board[self.pos]
			self.pos = np
			# update camera position
			self.cam = list(self.cam)
			for i in range(2):
				while self.pos[i] - self.cam[i] > 1:
					self.cam[i] += 1
				while self.cam[i] - self.pos[i] > 1:
					self.cam[i] -= 1
			self.cam = tuple(self.cam)
			# generate more beach
			while self.genmax < self.pos[1] + 10:
				self.genmax += 1
				self.genline(self.genmax)
			while self.genmin > self.pos[1] - 10:
				self.genmin -= 1
				self.genline(self.genmin)
		
		self.updateworld()
	
	def updateworld(self):
		self.movers = [self.randwalk(m) for m in self.movers]
		self.coastframe = (self.coastframe + 1) % 6
	
	def randwalk(self, m):
		# dancing crab
		if self.board[m].startswith('crab') and self.board[m] != 'crab':
			i = int(self.board[m][4:]) - 1
			if i:
				self.board[m] = 'crab' + str(i)
				return m
			else:
				self.board[m] = 'crab'
		
		dx = random.randint(-1, 1)
		dy = random.randint(-1, 1)
		np = (m[0] + dx, m[1] + dy)
		if np not in self.board:
			if np[0] > self.rockx and np[0] < self.coastx:
				self.board[np] = self.board[m]
				del self.board[m]
				return np
		elif np == self.pos:
			if self.board[m] == 'bird' \
			or (self.board[m] == 'crab' and self.board[self.pos] == 'baby'):
				global MENU
				MENU = Menu('you were eaten by a %s' % self.board[m])
				self.pos = None
				self.board[np] = self.board[m]
				del self.board[m]
				return np
		return m


def main():
	global GAME, MENU
	GAME = Game()
	MENU = Menu()
	pyglet.app.run()

main()

# vim: ts=4
